# -*- coding: utf-8 -*-
# Copyright (C) 2016 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of pytest-elasticsearch.

# pytest-elasticsearch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pytest-elasticsearch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with pytest-elasticsearch.  If not, see <http://www.gnu.org/licenses/>.
"""Pytest-elasticsearch py.test's plugin configuration."""
from pytest import Parser

from pytest_elasticsearch import factories

# pylint:disable=invalid-name
_help_host = "Elasticsearch host"
_help_executable = "Elasticsearch executable"
_help_port = "Elasticsearch port"
_help_cluster_name = "Cluster name of the elasticsearch process fixture"
_help_index_store_type = "type of the index to use in the elasticsearch process fixture"
_help_network_publish_host = "network host to which elasticsearch publish to connect to cluseter"
_help_elasticsearch_transport_tcp_port = "The tcp ansport port used \
    for internal communication between nodes within the cluster"


def pytest_addoption(parser: Parser) -> None:
    """Add plugin's configuration options."""
    parser.addini(name="elasticsearch_host", help=_help_host, default="127.0.0.1")

    parser.addini(
        name="elasticsearch_executable",
        help=_help_executable,
        default="/usr/share/elasticsearch/bin/elasticsearch",
    )

    parser.addini(name="elasticsearch_cluster_name", help=_help_cluster_name, default="")

    parser.addini(
        name="elasticsearch_index_store_type", help=_help_index_store_type, default="mmapfs"
    )

    parser.addini(
        name="elasticsearch_network_publish_host",
        help=_help_network_publish_host,
        default="127.0.0.1",
    )

    parser.addini(
        name="elasticsearch_port",
        help=_help_port,
        default=None,
    )

    parser.addini(
        "elasticsearch_transport_tcp_port",
        help=_help_elasticsearch_transport_tcp_port,
        default=None,
    )

    parser.addoption(
        "--elasticsearch-host",
        action="store",
        dest="elasticsearch_host",
        help=_help_host,
    )

    parser.addoption(
        "--elasticsearch-executable",
        action="store",
        dest="elasticsearch_executable",
        help=_help_executable,
    )

    parser.addoption(
        "--elasticsearch-cluster-name",
        action="store",
        dest="elasticsearch_cluster_name",
        help=_help_cluster_name,
    )

    parser.addoption(
        "--elasticsearch-index-store-type",
        action="store",
        dest="elasticsearch_index_store_type",
        help=_help_index_store_type,
    )

    parser.addoption(
        "--elasticsearch-network-publish-host",
        action="store",
        dest="elasticsearch_network_publish_host",
        help=_help_network_publish_host,
    )

    parser.addoption(
        "--elasticsearch-port", action="store", dest="elasticsearch_port", help=_help_port
    )

    parser.addoption(
        "--elasticsearch-transport-tcp-port",
        action="store",
        dest="elasticsearch_transport_tcp_port",
        help=_help_elasticsearch_transport_tcp_port,
    )


elasticsearch_proc = factories.elasticsearch_proc()
elasticsearch_nooproc = factories.elasticsearch_noproc()
elasticsearch = factories.elasticsearch("elasticsearch_proc")



# -*- coding: utf-8 -*-
# Copyright (C) 2016 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of pytest-elasticsearch.

# pytest-elasticsearch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pytest-elasticsearch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with pytest-elasticsearch.  If not, see <http://www.gnu.org/licenses/>.
"""Main packge for pytest-elasticsearch."""
__version__ = "4.0.0"



"""Configuration for pytest-elasticsearch."""
from pathlib import Path
from typing import Any, Optional, TypedDict

from pytest import FixtureRequest


class ElasticsearchConfigDict(TypedDict):
    """Typed Config dictionary."""

    executable: Path
    host: str
    port: Optional[int]
    transport_tcp_port: Optional[int]
    cluster_name: str
    network_publish_host: str
    index_store_type: str


def get_config(request: FixtureRequest) -> ElasticsearchConfigDict:
    """Return a dictionary with config options."""

    def get_elasticsearch_option(option: str) -> Any:
        name = "elasticsearch_" + option
        return request.config.getoption(name) or request.config.getini(name)

    return ElasticsearchConfigDict(
        executable=get_elasticsearch_option("executable"),
        host=get_elasticsearch_option("host"),
        port=get_elasticsearch_option("port"),
        transport_tcp_port=get_elasticsearch_option("transport_tcp_port"),
        cluster_name=get_elasticsearch_option("cluster_name"),
        network_publish_host=get_elasticsearch_option("network_publish_host"),
        index_store_type=get_elasticsearch_option("index_store_type"),
    )



"""Elasticsearch executor."""

import re
from pathlib import Path
from subprocess import check_output
from typing import Literal, Optional

from mirakuru import HTTPExecutor
from packaging.version import Version


class NoopElasticsearch:  # pylint:disable=too-few-public-methods
    """No operation Elasticsearch executor mock."""

    def __init__(self, host: str, port: int) -> None:
        """Initialize Elasticsearch executor mock.

        :param str host: hostname under which elasticsearch is available
        :param int port: port under which elasticsearch is available.
        """
        self.host = host
        self.port = port

    @staticmethod
    def running() -> Literal[True]:
        """Mock method pretending the executor is running."""
        return True


# pylint:disable=too-many-instance-attributes
class ElasticSearchExecutor(HTTPExecutor):
    """Elasticsearch executor."""

    def __init__(
        self,
        executable: Path,
        host: str,
        port: int,
        tcp_port: int,
        pidfile: Path,
        logs_path: Path,
        works_path: Path,
        cluster_name: str,
        network_publish_host: str,
        index_store_type: str,
        timeout: int,
    ) -> None:  # pylint:disable=too-many-arguments
        """Initialize ElasticSearchExecutor.

        :param executable: Executable path
        :param host: hostname under which elasticsearch will be running
        :param port: port elasticsearch listens on
        :param tcp_port: port used for internal communication
        :param pidfile: pidfile location
        :param logs_path: log files location
        :param works_path: workdir location
        :param cluster_name: cluster name
        :param network_publish_host: network host to which elasticsearch
            publish to connect to cluseter'
        :param index_store_type: type of the index to use in the
            elasticsearch process fixture
        :param timeout: Time after which to give up to start elasticsearch
        """
        self._version: Optional[Version] = None
        self.executable = executable
        self.host = host
        self.port = port
        # TODO: rename to transport_port
        self.tcp_port = tcp_port
        self.pidfile = pidfile
        self.logs_path = logs_path
        self.works_path = works_path
        self.cluster_name = cluster_name
        self.network_publish_host = network_publish_host
        self.index_store_type = index_store_type
        super().__init__(
            self._exec_command(),
            f"http://{self.host}:{self.port}",
            timeout=timeout,
        )

    @property
    def version(self) -> Version:
        """Get the given elasticsearch executable version parts.

        :return: Elasticsearch version
        """
        if not self._version:
            try:
                output = check_output([self.executable, "-Vv"]).decode("utf-8")
                match = re.search(r"Version: (?P<major>\d)\.(?P<minor>\d+)\.(?P<patch>\d+)", output)
                if not match:
                    raise RuntimeError(
                        "Elasticsearch version is not recognized. "
                        "It is probably not supported. \n"
                        "Output is: " + output
                    )
                version = match.groupdict()
                self._version = Version(
                    ".".join([version["major"], version["minor"], version["patch"]])
                )
            except OSError as exc:
                raise RuntimeError(
                    "'%s' does not point to elasticsearch." % self.executable
                ) from exc
        return self._version

    def _exec_command(self) -> str:
        """Get command to run elasticsearch binary based on the version.

        :return: command to run elasticsearch
        """
        port_param = "transport.port"
        if self.version < Version("7.0.0"):
            raise RuntimeError("This elasticsearch version is not supported.")
        elif self.version < Version("8.0.0"):
            port_param = "transport.tcp.port"
        else:
            port_param = "transport.port"
        return f"""
            {self.executable} -p {self.pidfile}
            -E http.port={self.port}
            -E {port_param}={self.tcp_port}
            -E path.logs={self.logs_path}
            -E path.data={self.works_path}
            -E cluster.name={self.cluster_name}
            -E network.host='{self.network_publish_host}'
            -E index.store.type={self.index_store_type}
            -E xpack.security.enabled=false
        """



# Copyright (C) 2013-2016 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of pytest-elasticsearch.

# pytest-elasticsearch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pytest-elasticsearch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with pytest-elasticsearch.  If not, see <http://www.gnu.org/licenses/>.
"""Fixture factories."""
import shutil
from pathlib import Path
from typing import Callable, Iterator, Optional

import pytest
from elasticsearch import Elasticsearch
from mirakuru import ProcessExitedWithError
from port_for import get_port
from port_for.api import PortType
from pytest import FixtureRequest, TempPathFactory

from pytest_elasticsearch.config import get_config
from pytest_elasticsearch.executor import ElasticSearchExecutor, NoopElasticsearch


def elasticsearch_proc(
    executable: Optional[Path] = None,
    host: Optional[str] = None,
    port: Optional[PortType] = -1,
    transport_tcp_port: Optional[PortType] = None,
    cluster_name: Optional[str] = None,
    network_publish_host: Optional[str] = None,
    index_store_type: Optional[str] = None,
) -> Callable[[FixtureRequest, TempPathFactory], Iterator[ElasticSearchExecutor]]:
    """Create elasticsearch process fixture.

    :param executable: elasticsearch's executable
    :param host: host that the instance listens on
    :param port:
        exact port (e.g. '8000', 8000)
        randomly selected port (None) - any random available port
        [(2000,3000)] or (2000,3000) - random available port from a given range
        [{4002,4003}] or {4002,4003} - random of 4002 or 4003 ports
        [(2000,3000), {4002,4003}] -random of given range and set
    :param transport_tcp_port: Port used for communication between nodes
    :param cluster_name: name of a cluser this node should work on.
        Used for autodiscovery. By default each node is in it's own cluser.
    :param network_publish_host: host to publish itself within cluser
        http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/modules-network.html
    :param index_store_type: index.store.type setting. *memory* by default
        to speed up tests
    """

    @pytest.fixture(scope="session")
    def elasticsearch_proc_fixture(
        request: FixtureRequest, tmp_path_factory: TempPathFactory
    ) -> Iterator[ElasticSearchExecutor]:
        """Elasticsearch process starting fixture."""
        config = get_config(request)
        elasticsearch_host = host or config["host"]
        elasticsearch_executable = executable or config["executable"]

        elasticsearch_port = get_port(port) or get_port(config["port"])
        assert elasticsearch_port
        elasticsearch_transport_port = get_port(
            transport_tcp_port, exclude_ports=[elasticsearch_port]
        ) or get_port(config["transport_tcp_port"], exclude_ports=[elasticsearch_port])
        assert elasticsearch_transport_port

        elasticsearch_cluster_name = (
            cluster_name or config["cluster_name"] or f"elasticsearch_cluster_{elasticsearch_port}"
        )
        assert elasticsearch_cluster_name
        elasticsearch_index_store_type = index_store_type or config["index_store_type"]
        elasticsearch_network_publish_host = network_publish_host or config["network_publish_host"]
        tmpdir = tmp_path_factory.mktemp(f"pytest-elasticsearch-{request.fixturename}")

        logs_path = tmpdir / "logs"

        pidfile = tmpdir / f"elasticsearch.{elasticsearch_port}.pid"
        work_path = tmpdir / f"workdir_{elasticsearch_port}"

        elasticsearch_executor = ElasticSearchExecutor(
            elasticsearch_executable,
            elasticsearch_host,
            elasticsearch_port,
            elasticsearch_transport_port,
            pidfile,
            logs_path,
            work_path,
            elasticsearch_cluster_name,
            elasticsearch_network_publish_host,
            elasticsearch_index_store_type,
            timeout=60,
        )
        print(elasticsearch_executor.command)

        elasticsearch_executor.start()
        yield elasticsearch_executor
        try:
            elasticsearch_executor.stop()
        except ProcessExitedWithError:
            pass
        shutil.rmtree(work_path)
        shutil.rmtree(logs_path)

    return elasticsearch_proc_fixture


def elasticsearch_noproc(
    host: Optional[str] = None, port: Optional[int] = None
) -> Callable[[FixtureRequest], Iterator[NoopElasticsearch]]:
    """Elasticsearch noprocess factory.

    :param host: hostname
    :param port: exact port (e.g. '8000', 8000)
    :returns: function which makes a elasticsearch process
    """

    @pytest.fixture(scope="session")
    def elasticsearch_noproc_fixture(request: FixtureRequest) -> Iterator[NoopElasticsearch]:
        """Noop Process fixture for PostgreSQL.

        :param FixtureRequest request: fixture request object
        :rtype: pytest_dbfixtures.executors.TCPExecutor
        :returns: tcp executor-like object
        """
        config = get_config(request)
        es_host = host or config["host"]
        assert es_host
        es_port = port or config["port"] or 9300
        assert es_port

        yield NoopElasticsearch(host=es_host, port=es_port)

    return elasticsearch_noproc_fixture


def elasticsearch(process_fixture_name: str) -> Callable[[FixtureRequest], Iterator[Elasticsearch]]:
    """Create Elasticsearch client fixture.

    :param process_fixture_name: elasticsearch process fixture name
    """

    @pytest.fixture
    def elasticsearch_fixture(request: FixtureRequest) -> Iterator[Elasticsearch]:
        """Elasticsearch client fixture."""
        process = request.getfixturevalue(process_fixture_name)
        if not process.running():
            process.start()
        client = Elasticsearch(
            hosts=[{"host": process.host, "port": process.port, "scheme": "http"}],
            request_timeout=30,
            verify_certs=False,
        )
        client.options(ignore_status=400)

        yield client
        for index in client.indices.get_alias():
            client.indices.delete(index=index)

    return elasticsearch_fixture
